/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 */
#include "runtime.h"
#include "list.h"

struct node {
	void *data;
	int deleted;
	struct node *next,*prev;
};

struct list {
	POOL pool;
	struct node *head,*tail;
	struct node *current;
	list_delete_func_t delete_func;
};

struct list *list_create(POOL p)
{
	POOL lpool;
	struct list *list;
	lpool = pool_new(p);
	list = XMALLOC(lpool,struct list);
	list->pool = lpool;

	return list;
}
int list_destroy(struct list *list)
{
	ASSERT(list);
	/* Fastest damn free you ever saw */
	pool_destroy(list->pool);
	return 0;
}

/* Calls callback for each item in the list.
 * Just before calling callback, sets the current pointer to that
 * node.
 */
int list_foreach(struct list *list, list_callback_t callback, void *userdata)
{
	struct node *n,*o;
	ASSERT(list);
	ASSERT(callback);
	n = list->head;
	/* Careful, user can legally delete the current node from under
	 * our feet.
	 */
	while(n != NULL) {
		o = n;
		n=n->next;
		if (o->deleted)
			continue;
		list->current = o;
		if (callback(list,o->data,userdata) == -1)
			break;
	}
	return n ? -1 : 0;
}

int list_foreach_simple(struct list *list, list_simple_callback_t callback)
{
	struct node *n,*o;
	ASSERT(list);
	ASSERT(callback);
	n = list->head;
	/* Careful, user can legally delete the current node from under
	 * our feet.
	 */
	while(n != NULL) {
		o = n;
		n=n->next;
		if (o->deleted)
			continue;
		list->current = o;
		callback(o->data);
	}
	return 0;

}
/* Adds an item to the beginning of the list.
 * Sets current pointer
 */
int list_add_front(struct list *list, void *contents)
{
	struct node *n;
	ASSERT(list);
	n = XMALLOC(list->pool,struct node);
	n->data = contents;
	n->prev = NULL;
	n->next = list->head;
	if (list->head)
		list->head->prev = n;
	list->head = n;
	if (!list->tail)
		list->tail = n;
	return 0;
}
/* Adds an item to the end of the list
 * Sets current pointer
 */
int list_add_end(struct list *list, void *contents)
{
	struct node *n;
	ASSERT(list);
	n = XMALLOC(list->pool,struct node);
	n->data = contents;
	n->prev = list->tail;
	n->next = NULL;
	if (list->tail)
		list->tail->next = n;
	list->tail = n;
	if (!list->head)
		list->head = n;
	return 0;
}

/* finds an item in the list, calling comp_func(what,item) for each
 * item in the list until comp_func returns 0, then returns item.
 * returns NULL if not found.
 * Sets current pointer
 */
void *list_find(struct list *list, void *what, list_compare_t comp_func)
{
	struct node *n;
	ASSERT(list);
	ASSERT(comp_func);

	for(n=list->head; n != NULL; n=n->next) {
		if (!n->deleted && comp_func(what,n->data) == 0) {
			list->current = n;
			return n->data;
		}
	}
	return NULL;
}

void *list_find_data(struct list *list, void *data)
{
	struct node *n;
	ASSERT(list);

	for(n=list->head; n != NULL; n=n->next) {
		if (!n->deleted && data == n->data) {
			list->current = n;
			return n->data;
		}
	}
	return NULL;
}
void *list_get_first(struct list *list)
{
	void *d;

	if (!list->head)
		return NULL;
	d = list->head->data;
	list->current = list->head;
	return d;
}

void *list_pop_first(struct list *list)
{
	void *d;

	if (!list->head)
		return NULL;
	d = list->head->data;
	list->current = list->head;
	list_del_current(list);
	return d;
}
void *list_pop_last(struct list *list)
{
	void *d;

	if (!list->tail)
		return NULL;
	d = list->tail->data;
	list->current = list->tail;
	list_del_current(list);
	return d;
}

/* Deletes the node at the current pointer. (Set with list_foreach or
 * list_find).
 * Sets current pointer to NULL afterwards.
 * Returns -1 if current pointer was already NULL.
 */
int list_del_current(struct list *list)
{
	ASSERT(list);
	if (list->current == NULL)
		return -1;
	if (list->current->prev)
		list->current->prev->next = list->current->next;
	else
		list->head = list->current->next;
	if (list->current->next)
		list->current->next->prev = list->current->prev;
	else
		list->tail = list->current->prev;
	list->current = NULL;
	return 0;
}

int list_delete_func(struct list *list, list_delete_func_t callback)
{
	list->delete_func = callback;
	return 0;
}

/* Calls the list delete function for every node that has been marked
 * for deletion (by a call to the list_mark_current() function.
 */
int list_cleanup(struct list *list)
{
	struct node *n,*o;
	ASSERT(list);
	if (list->delete_func == NULL)
		return -1;
	n = list->head;
	/* Careful, user can legally delete the current node from under
	 * our feet.
	 */
	while(n != NULL) {
		o = n;
		n=n->next;
		if (!o->deleted)
			continue;
		list->current = o;
		if (list->delete_func(list,o->data) == -1)
			break;
		if (list->current == NULL) {
			/* user has called delete_current, so skip this
			 */
			continue;
		}
		list_del_current(list);
	}
	return n ? -1 : 0;
}

/* Marks the current node for deletion when the list_cleanup function
 * is called.
 */
int list_mark_current(struct list *list)
{
	ASSERT(list);
	if (list->current == NULL)
		return -1;
	list->current->deleted = 1;
	return 0;
}

